# Examples: 3-D Subduction Zone

This suite of examples demonstrates use of a wide variety of features
and the general workflow often used in research simulations.  The
examples consistent of a step-by-step sequence of eight problems
involving a 3D subduction zone. They focus on modeling the
deformation associated with the the subducting slab, including
interseismic deformation with aseismic slip (creep) and viscoelastic
relaxation, coseismic slip on the slab interface and a splay fault,
and slow slip events on the subduction interface. We want to account
for the 3D material properties associated with different elastic
properties for the subducting slab, mantle, continental crust, and an
accretionary wedge. All of the examples use the same mesh, generated
using Cubit.

Directory structure:
* **input**: Large input files not included in PyLith (mesh)
* **output**: Simulation output files, created automatically by PyLith
* **scratch**: Files generated by Python scripts.
* **utils**: Utilities used to construct input files
* **viz**: ParaView Python scripts

Examples:

* `Step 1`: Static axial compression
* `Step 2`: Quasistatic coseismic and postseismic deformation from an earthquake rupture in the center of the subduction zone interface
* `Step 3`: Quasi-static interseisic deformation with creep on the top and bottom of the slab
* `Step 4`: Quasi-static earthquake cycle with prescribed earthquake rupture and creep
* `Step 5`: Spontaneous rupture driven by subducting slab (not yet updated for v3.0)
* `Step 6`: Prescribed slow-slip event (not yet updated for v3.0)
* `Step 7a,b`: Inversion of slow-slip event using 3D Green's functions (not yet updated for v3.0)
* `Step 8a,b,c`: Stress field due to gravitational body forces (not yet updated for v3.0)

## Mesh generation using Cubit (optional)

We use Cubit to generate the finite-element mesh. Due to its size, we do not include the finite-element mesh in the PyLith source or binary distributions. If you do not have Cubit,
you can download the mesh from
https://github.com/geodynamics/pylith/releases/download/v3.0.1/mesh_tet.exo.gz and skip
generating the mesh. See `input/README.md` for instructions on how to download the file.

```
# Run the Python script to generate the journal file 'scratch/cubit_create_surfs.jou'.
./utils/generate_surfjou.py
```

The next step is to use Cubit to run the `cubit_create_surfs.jou` journal file to generate the spline surfaces for the slab and splay fault and save them as ACIS surfaces.

After you generate the ACIS surface files, run the `cubit_tet.jou` journal file to construct the geometry, and generate the mesh. In the end you will have an Exodus-II file `mesh_tet.exo`, which is a NetCDF file, in the `input` directory.

## Step 1: Axial Compression

We start with a very simple example of axial compression in the east-west (x) direction with purely elastic material properties, and no faults.

To run this example:
```
pylith step01_axialdisp.cfg mat_elastic.cfg
```

## Step 2: Earthquake Rupture and Postseismic Relaxation

In this example we model the postseismic relaxation of the deep slab and mantle resulting from
prescribed coseismic slip on a fault patch in the central portion of the subduction (top of the 
slab) interface.

To run this example:
```
pylith step02_coseismic.cfg mat_viscoelastic.cfg
```

## Step 3: Prescribed Aseismic Creep and Interseismic Deformation

We increase the complexity of our fault model by simulating the interseismic deformation associated 
with the subducting slab.

To run this example:
```
pylith step93_interseismic.cfg mat_viscoelastic.cfg
```

## Step 4: Prescribed Earthquake Cycle

In Step 4 we combine the interseismic deformation in Step 3 with the coseismic slip in Step 2 to 
simulate two earthquake cycles. We also include an earthquake on the splay fault.

To run this example:
```
pylith step04_eqcycle.cfg mat_viscoelastic.cfg
```

## Step 5: Spontaneous Rupture Driving by Subducting Slab

This example is not yet implemented.

## Step 6: Prescribed Slow-Slip Event

This static simulation simulates a simple slow slip event (SSE) on the subduction interface, in 
which the entire patch slips simultaneously with an amplitude that grows with time. 

To run this example:
```
# Generate the spatial database files: 
#   fault_slabtop_slowslip.spatialdb
#   fault_slabtop_slowslip.timedb
./utils/generate_slowslip.py
pylith step06_slowslip.cfg mat_elastic.cfg
```

## Step 7: Inversion of Slow-Slip Event using 3D Green's Functions

**Warning**: This example is still under construction.

We compute the static Green's function for prescribed slip imulses on the central portion of the 
subduction interface. We then use the responses to perform a simple geodetic inversion of the slip 
in Step 7. We divide generating Green's functions into two sub-problems:

* **Step 7a**: Left-lateral slip component
* **Step 7b**: Reverse slip component

Although PyLith can generate the two components in one simulation, we often prefer to speed up the 
process by running simulations for each of the components at the same time using multiple processes 
on a cluster.

To run this example:
```
pylith step07a_leftlateral.cfg mat_elastic.cfg
pylith step07b_reverse.cfg mat_elastic.cfg
```

Before we run the inversion, we postprocess the output from Step 6 to create synthetic data. 
```
# Generate synthetic GPS data
./utils/make_synthetic_gpsdisp.py
```

We perform a simple inversion using the `slip_invert.py` Python script.
```
./utils/slip_invert.py
```

## Step 8: Stress Field Due to Gravitational Body Forces

**Warning**: This example is still under construction.

This example demonstrates the use of gravitational body forces as well
as the use of initial stresses to balance the body forces. This
involves enabling gravity within our domain with Dirichlet roller
boundary conditions on the lateral and bottom boundaries; we do not
include faults in this example.  We also demonstrate what happens when
the initial stresses are not in balance with the gravitational
stresses, and show how viscoelastic problems with gravitational
stresses will in general not reach a steady-state solution. The
example is divided into three sub-problems:

  * **Step 8a**: Gravitational body forces with 3-D density variations
    in elastic materials and initial stresses for a uniform density.
  * **Step 8b**: Gravitational body forces with 3-D density variations
    in incompressible elastic materials.
  * **Step 8c**: Gravitational body forces with 3-D density variations
    in elastic and viscoelastic materials and initial stresses from
    Step 8a plus finite strain formulation (does not reach a
    steady-state solution).

To run this example:
```
# Step 8a
pylith step 8a_gravity_refstate.cfg mat_elastic.cfg solver_algebraicmultigrid.cfg

# Step 8b
pylith step 8b_gravity_incompressible.cfg mat_incompressible.cfg solver_algebraicmultigrid.cfg

# Step 8c
pylith step 8c_gravity_viscoelastic.cfg mat_viscoelastic.cfg solver_algebraicmultigrid.cfg
```

\end{shell}

