# =================================================================================================
# This code is part of PyLith, developed through the Computational Infrastructure
# for Geodynamics (https://github.com/geodynamics/pylith).
#
# Copyright (c) 2010-2023, University of California, Davis and the PyLith Development Team.
# All rights reserved.
#
# See https://mit-license.org/ and LICENSE.md and for license information. 
# =================================================================================================

from pylith.utils.PetscComponent import PetscComponent


class AuxSubfieldsElasticity(PetscComponent):
    """
    Auxiliary subfields associated with the elasticity equation.

    Setting the parameters for a subfield does not turn on its use.
    The [`Elasticity` Component](Elasticity.md) has flags for including or excluding terms in the elasticity equation.
    """
    DOC_CONFIG = {
        "cfg": """
            # We set the basis order to represent linear variations in the density and body 
            # force subfields and a uniform gravitational acceleration subfield.
            [pylithapp.problem.materials.mat_elastic.auxiliary_fields]
            density.basis_order = 1
            body_force.basis_order = 1
            gravitational_acceleration.basis_order = 0
        """
    }

    import pythia.pyre.inventory
    from pylith.topology.Subfield import Subfield

    density = pythia.pyre.inventory.facility("density", family="auxiliary_subfield", factory=Subfield)
    density.meta['tip'] = "Density subfield."

    bodyForce = pythia.pyre.inventory.facility("body_force", family="auxiliary_subfield", factory=Subfield)
    bodyForce.meta['tip'] = "Body force subfield."

    gravitationalAcceleration = pythia.pyre.inventory.facility("gravitational_acceleration", family="auxiliary_subfield", factory=Subfield)
    gravitationalAcceleration.meta['tip'] = "Gravitational acceleration subfield."

    def __init__(self, name="auxsubfieldselasticity"):
        """Constructor.
        """
        PetscComponent.__init__(self, name, facility="auxiliary_subfields")

    def _configure(self):
        PetscComponent._configure(self)


# FACTORIES ////////////////////////////////////////////////////////////

def auxiliary_subfields():
    """Factory associated with AuxSubfieldsElasticity.
    """
    return AuxSubfieldsElasticity()


# End of file
